@@ -34,11 +34,11 @@ type zresult = {
 }
 
 external zlib_deflate_init2 : int -> int -> zstream = "zlib_deflate_init2"
-external zlib_deflate : zstream -> src:string -> spos:int -> slen:int -> dst:string -> dpos:int -> dlen:int -> zflush -> zresult = "zlib_deflate_bytecode" "zlib_deflate"
+external zlib_deflate : zstream -> src:string -> spos:int -> slen:int -> dst:bytes -> dpos:int -> dlen:int -> zflush -> zresult = "zlib_deflate_bytecode" "zlib_deflate"
 external zlib_deflate_end : zstream -> unit = "zlib_deflate_end"
 
 external zlib_inflate_init2 : int -> zstream = "zlib_inflate_init"
-external zlib_inflate : zstream -> src:string -> spos:int -> slen:int -> dst:string -> dpos:int -> dlen:int -> zflush -> zresult = "zlib_inflate_bytecode" "zlib_inflate"
+external zlib_inflate : zstream -> src:string -> spos:int -> slen:int -> dst:bytes -> dpos:int -> dlen:int -> zflush -> zresult = "zlib_inflate_bytecode" "zlib_inflate"
 external zlib_inflate_end : zstream -> unit = "zlib_inflate_end"
 
 external _executable_path : string -> string = "executable_path"
@@ -47,6 +47,8 @@ external get_real_path : string -> string = "get_real_path"
 
 external zlib_deflate_bound : zstream -> int -> int = "zlib_deflate_bound"
 
+external zlib_crc32 : bytes -> int -> int32 = "zlib_crc32"
+
 external time : unit -> float = "sys_time"
 
 type library
@@ -91,26 +93,26 @@ let executable_path() =
 
 let zlib_op op z str =
 	let bufsize = 1 lsl 14 in
-	let tmp = String.create bufsize in
+	let tmp = Bytes.create bufsize in
 	let total = ref 0 in
 	let rec loop pos len acc =
 		let r = op z ~src:str ~spos:pos ~slen:len ~dst:tmp ~dpos:0 ~dlen:bufsize (if len = 0 then Z_FINISH else Z_SYNC_FLUSH) in
 		total := !total + r.z_wrote;
-		let acc = String.sub tmp 0 r.z_wrote :: acc in
+		let acc = Bytes.sub tmp 0 r.z_wrote :: acc in
 		if r.z_finish then
 			acc
 		else
 			loop (pos + r.z_read) (len - r.z_read) acc
 	in
 	let strings = loop 0 (String.length str) [] in
-	let big = String.create !total in
+	let big = Bytes.create !total in
 	ignore(List.fold_left (fun p s ->
-		let l = String.length s in
+		let l = Bytes.length s in
 		let p = p - l in
-		String.unsafe_blit s 0 big p l;
+		Bytes.unsafe_blit s 0 big p l;
 		p
 	) !total strings);
-	big
+	Bytes.unsafe_to_string big
 
 let zip str =
 	let z = zlib_deflate_init 9 in
@@ -125,8 +127,8 @@ let unzip str =
 	s
 
 let input_zip ?(bufsize=65536) ch =
-	let tmp_out = String.create bufsize in
-	let tmp_in = String.create bufsize in
+	let tmp_out = Bytes.create bufsize in
+	let tmp_in = Bytes.create bufsize in
 	let tmp_buf = Buffer.create bufsize in
 	let buf = ref "" in
 	let p = ref 0 in
@@ -134,8 +136,8 @@ let input_zip ?(bufsize=65536) ch =
 	let rec fill_buffer() =
 		let rec loop pos len =
 			if len > 0 || pos = 0 then begin
-				let r = zlib_inflate z tmp_in pos len tmp_out 0 bufsize (if pos = 0 && len = 0 then Z_FINISH else Z_SYNC_FLUSH) in
-				Buffer.add_substring tmp_buf tmp_out 0 r.z_wrote;
+				let r = zlib_inflate z (Bytes.unsafe_to_string tmp_in) pos len tmp_out 0 bufsize (if pos = 0 && len = 0 then Z_FINISH else Z_SYNC_FLUSH) in
+				Buffer.add_subbytes tmp_buf tmp_out 0 r.z_wrote;
 				loop (pos + r.z_read) (len - r.z_read);
 			end
 		in
@@ -172,30 +174,30 @@ let input_zip ?(bufsize=65536) ch =
 
 let output_zip ?(bufsize=65536) ?(level=9) ch =
 	let z = zlib_deflate_init level in
-	let out = String.create bufsize in
-	let tmp_out = String.create bufsize in
+	let out = Bytes.create bufsize in
+	let tmp_out = Bytes.create bufsize in
 	let p = ref 0 in
 	let rec flush finish =
-		let r = zlib_deflate z out 0 !p tmp_out 0 bufsize (if finish then Z_FINISH else Z_SYNC_FLUSH) in
+		let r = zlib_deflate z (Bytes.unsafe_to_string out) 0 !p tmp_out 0 bufsize (if finish then Z_FINISH else Z_SYNC_FLUSH) in
 		ignore(IO.really_output ch tmp_out 0 r.z_wrote);
 		let remain = !p - r.z_read in
-		String.blit out r.z_read out 0 remain;
+		Bytes.blit out r.z_read out 0 remain;
 		p := remain;
 		if finish && not r.z_finish then flush true
 	in
 	let write c =
 		if !p = bufsize then flush false;
-		String.unsafe_set out !p c;
+		Bytes.unsafe_set out !p c;
 		incr p
 	in
 	let rec output str pos len =
 		let b = bufsize - !p in
 		if len <= b then begin
-			String.blit str pos out !p len;
+			Bytes.blit str pos out !p len;
 			p := !p + len;
 			len
 		end else begin
-			String.blit str pos out !p b;
+			Bytes.blit str pos out !p b;
 			p := !p + b;
 			flush false;
 			b + output str (pos + b) (len - b);
