@@ -1,6 +1,6 @@
 (*
 	The Haxe Compiler
-	Copyright (C) 2005-2017  Haxe Foundation
+	Copyright (C) 2005-2019  Haxe Foundation
 
 	This program is free software; you can redistribute it and/or
 	modify it under the terms of the GNU General Public License
@@ -70,7 +70,6 @@ module BasicBlock = struct
 		| SEWhile of t * t * t                                   (* `while` with "head", "body" and "next" *)
 		| SESubBlock of t * t                                    (* "sub" with "next" *)
 		| SEMerge of t                                           (* Merge to same block *)
-		| SEEnd                                                  (* End of syntax *)
 		| SENone                                                 (* No syntax exit *)
 
 	and t = {
@@ -211,17 +210,18 @@ module Graph = struct
 		} in
 		DynArray.add g.g_var_infos vi;
 		let i = DynArray.length g.g_var_infos - 1 in
-		v.v_extra <- Some([],Some (mk (TConst (TInt (Int32.of_int i))) t_dynamic null_pos))
+		v.v_extra <- Some([],Some (mk (TConst (TInt (Int32.of_int i))) t_dynamic null_pos));
+		vi
 
 	let get_var_info g v = match v.v_extra with
 		| Some(_,Some {eexpr = TConst (TInt i32)}) -> DynArray.get g.g_var_infos (Int32.to_int i32)
 		| _ ->
-			prerr_endline "Unbound variable, please report this";
-			prerr_endline (Printer.s_tvar v);
-			assert false
+			print_endline "Unbound variable, please report this";
+			print_endline (Printer.s_tvar v);
+			create_var_info g g.g_unreachable v
 
 	let declare_var g v bb =
-		create_var_info g bb v
+		ignore(create_var_info g bb v)
 
 	let add_var_def g bb v =
 		if bb.bb_id > 0 then begin
@@ -311,15 +311,15 @@ module Graph = struct
 		List.iter (fun bb ->
 			List.iter (fun edge ->
 				if edge.cfg_to = g.g_unreachable then
-					prerr_endline (Printf.sprintf "Outgoing edge from %i to the unreachable block" bb.bb_id)
+					print_endline (Printf.sprintf "Outgoing edge from %i to the unreachable block" bb.bb_id)
 				else if not (List.memq edge edge.cfg_to.bb_incoming) then
-					prerr_endline (Printf.sprintf "Outgoing edge %i -> %i has no matching incoming edge" edge.cfg_from.bb_id edge.cfg_to.bb_id)
+					print_endline (Printf.sprintf "Outgoing edge %i -> %i has no matching incoming edge" edge.cfg_from.bb_id edge.cfg_to.bb_id)
 			) bb.bb_outgoing;
 			List.iter (fun edge ->
 				if edge.cfg_from == g.g_unreachable then
-					prerr_endline (Printf.sprintf "Incoming edge to %i from the unreachable block" bb.bb_id)
+					print_endline (Printf.sprintf "Incoming edge to %i from the unreachable block" bb.bb_id)
 				else if not (List.memq edge edge.cfg_from.bb_outgoing) then
-					prerr_endline (Printf.sprintf "Incoming edge %i <- %i has no matching outgoing edge" edge.cfg_to.bb_id edge.cfg_from.bb_id)
+					print_endline (Printf.sprintf "Incoming edge %i <- %i has no matching outgoing edge" edge.cfg_to.bb_id edge.cfg_from.bb_id)
 			) bb.bb_incoming
 		) g.g_nodes
 
@@ -457,10 +457,10 @@ module Graph = struct
 					()
 			end;
 			DynArray.iter (fun e -> match e.eexpr with
-				| TVar(v,eo) when not (is_unbound v) ->
+				| TVar(v,eo) ->
 					declare_var g v bb;
 					if eo <> None then add_var_def g bb v;
-				| TBinop(OpAssign,{eexpr = TLocal v},_) when not (is_unbound v) ->
+				| TBinop(OpAssign,{eexpr = TLocal v},_) ->
 					add_var_def g bb v
 				| _ ->
 					()
@@ -505,7 +505,7 @@ module Graph = struct
 					loop scopes bb_next
 				| SEMerge bb ->
 					loop scopes bb
-				| SEEnd | SENone ->
+				| SENone ->
 					()
 			end
 		in
@@ -523,4 +523,4 @@ type analyzer_context = {
 	mutable loop_stack : int list;
 	mutable debug_exprs : (string * texpr) list;
 	mutable name_stack : string list;
-}
\ No newline at end of file
+}
